#!/usr/bin/python
'''Intended to be the one CLI bluecat tool to rule them all
'''
import axapi
import argparse
import json
import logging
import os
import requests
from ipaddress import ip_address, ip_network
from pybluecat import BAM
from pybluecat import data as DATA
from pybluecat.exceptions import BluecatError
from time import sleep


def get_creds(args):
    """Load credentials from file if given
    pull in from environment variables otherwise"""
    if args.creds is not None:
        # Load creds from file
        with open(args.creds) as f:
            creds = json.load(f)
    else:
        # Load creds from Environment variables
        default_filepath = os.environ['HOME'] + '/.bluecat'
        with open(default_filepath) as f:
            creds = json.load(f)
    return creds


def get_client(loglevel):
    """Instantiate bluecat client"""
    hostname = creds['bluecat']['hostname']
    username = creds['bluecat']['username']
    password = creds['bluecat']['password']
    bluecat = BAM(hostname, username, password, loglevel=loglevel)
    return bluecat


def format_mac(mac):
    """Returns a MAC address sans delimiters"""
    return mac.replace('.', '').replace(':', '').replace('-', '')


def process_dhcp_csv(filepath):
    """Converts a DHCP CSV file into a list of dicts"""
    with open(filepath) as f:
        csv = [line.strip().split(',') for line in f.readlines()]
    if any(field in csv[0] for field in ['hostname', 'mac', 'macaddress', 'network']):
        csv.pop(0)
    if '' in csv[-1]:
        csv.pop(-1)
    device_list = [{
            'name': device[0],
            'mac': device[1],
            'net': device[2].split('/')[0]
        }
        for device in csv
    ]
    return device_list


def find_mac_in_net(mac, network):
    """Checks for a MAC address within a network
    return the reservation object if found, None otherwise"""
    mac = format_mac(mac)
    entity = bluecat.get_mac_address(mac)
    try:
        results = bluecat.get_linked_entities(entity['id'])
    except BluecatError:
        return None
    # if 'Object was not found' not in results:
    for result in results:
        ip = bluecat.entity_to_json(result)
        if ip_address(ip['properties']['address']) in network:
            return ip
    return None

def find_name_in_net(name, network):
    """Checks for a reservation with for the given name in the  given
    network. Returns the object if found, otherwise: None"""
    search_results = bluecat.search_ip_by_name(name)
    for result in search_results:
        ip = bluecat.entity_to_json(result)
        if ip_address(ip['properties']['address']) in network:
            if name == ip['name']:
                return ip
    return None


def calculate_offset(network):
    if network.prefixlen > 24:
        offset = None
    else:
        offset = str(network.network_address + 31)
    return offset


def create_static(args):
    pass


def delete_static(args):
    pass


def update_static(args):
    pass


def create_static_bulk(device_list):
    pass


def delete_static_bulk(device_list):
    pass


def update_static_bulk(device_list):
    pass


def create_dhcp(args):
    # determine which networks to use
    if args.network:
        networks = [ip_network(unicode(args.network))]
    else:
        networks = lookup_enviornment_or_some_shit(args.environment)
    # loop through networks until reservation is made
    output = {}
    server_set = set()
    for network in networks:
        # Get Network info from Bluecat
        net_obj = bluecat.get_network(str(network.network_address))
        net_obj = bluecat.entity_to_json(net_obj)
        net = ip_network(net_obj['properties']['CIDR'])
        # Check for existing reservations
        reservation = find_mac_in_net(args.mac, net)
        if reservation is not None:
            output = reservation
            break
        # Set offset for network
        offset = calculate_offset(net)
        # Assign Next Available IP
        mac = format_mac(args.mac)
        response = bluecat.assign_next_ip_address(net_obj['id'], args.hostname, macAddr=mac, action='MAKE_DHCP_RESERVED', offset=offset)
        if bluecat.history[-1].status_code == 200:
            server_set = queue_servers(server_set, net_obj['id'])
            output = bluecat.entity_to_json(response)
            dns_response = create_dns_a_record(args.hostname, output['properties']['address'])
            output['dns-status'] = dns_response.json()['permalink']
            break
    deploy_dhcp_and_monitor(server_set)
    return output


def delete_dhcp(args):
    server_set = set()
    mac_entity = bluecat.get_mac_address(args.mac)
    linked_entities = bluecat.get_linked_entities(mac_entity['id'])
    for ip in linked_entities:
        ip = bluecat.entity_to_json(ip)
        net = bluecat.get_network(ip['properties']['address'])
        response = queue_servers(server_set, net['id'])



def update_dhcp(args):
    pass


def handle_dhcp_bulk(args):
    """Single function to dole out bulk dhcp actions"""
    device_list = process_dhcp_csv(args.filepath)
    if args.delete:
        output = delete_dhcp_bulk(device_list)
    elif args.update:
        output = update_dhcp_bulk(device_list)
    else:
        output = create_dhcp_bulk(device_list)
    return output


def create_dhcp_bulk(device_list):
    output = []
    server_set = set()
    for device in device_list:
        # Get Network info from Bluecat
        net_obj = bluecat.get_network(device['net'])
        net_obj = bluecat.entity_to_json(net_obj)
        net = ip_network(net_obj['properties']['CIDR'])
        # Set offset for network
        offset = calculate_offset(net)
        # Check for existing reservations
        reservation = find_mac_in_net(device['mac'], net)
        if reservation is not None:
            output.append(reservation)
            continue
        else:
            mac = format_mac(device['mac'])
            response = bluecat.assign_next_ip_address(net_obj['id'], device['name'], device['mac'], action='MAKE_DHCP_RESERVED', offset=offset)
            server_set = queue_servers(server_set, net_obj['id'])
            output.append(bluecat.entity_to_json(response))
            # ip_obj = bluecat.entity_to_json(response)
            dns_response = create_dns_a_record(device['name'], response['properties']['address'])
            output[-1]['dns-status'] = dns_response.json()['permalink']
    deploy_dhcp_and_monitor(server_set)
    return output


def delete_dhcp_bulk(device_list):
    pass


def update_dhcp_bulk(device_list):
    output = []
    server_set = set()
    for device in device_list:
        output_entry = {'deleted':[], 'created':[]}
        # Get Network info from Bluecat
        net_obj = bluecat.get_network(device['net'])
        net_obj = bluecat.entity_to_json(net_obj)
        net = ip_network(net_obj['properties']['CIDR'])
        # Find other instances of device of same name
        search_results = bluecat.search_ip_by_name(device['name'])
        for result in search_results:
            # Ensure we only delete exact matches, excluding case
            if result['name'].lower() == device['name'].lower():
                reservation = bluecat.entity_to_json(result)
                # del_addr = ip_address(reservation['properties']['address'])
                del_net = bluecat.get_network(reservation['properties']['address'])
                # delete the reservation from Bluecat
                bluecat.delete(reservation['id'])
                # delete the dns record
                dns_response = delete_dns_a_record(reservation['name'].lower())
                # ensure the changes will get deployed to the correct servers
                server_set = queue_servers(server_set, del_net['id'])
                # log the deletion
                reservation['dns-status'] = dns_response.json()['permalink']
                output_entry['deleted'].append(reservation)
        # Set offset for network
        offset = calculate_offset(net)
        # Check for existing reservations
        reservation = find_mac_in_net(device['mac'], net)
        if reservation is not None:
            output_entry['created'].append(reservation)
            continue
        else:
            mac = format_mac(device['mac'])
            response = bluecat.assign_next_ip_address(net_obj['id'], device['name'], device['mac'], action='MAKE_DHCP_RESERVED', offset=offset)
            server_set = queue_servers(server_set, net_obj['id'])
            output_entry['created'].append(bluecat.entity_to_json(response))
            # ip_obj = bluecat.entity_to_json(response)
            dns_response = create_dns_a_record(device['name'], response['properties']['address'])
            output_entry['created'][-1]['dns-status'] = dns_response.json()['permalink']
        output.append(output_entry)
    deploy_dhcp_and_monitor(server_set)
    return output


def queue_servers(server_set, network_id):
    roles = bluecat.get_deployment_roles(network_id)
    server_primary = bluecat.get_server_for_role(roles[0]['id'])
    server_backup_id = DATA.ADONIS_PAIRS[server_primary['id']]
    server_set.add(server_primary['id'])
    server_set.add(server_backup_id)
    return server_set


def deploy_dhcp_and_monitor(server_set):
    for server in server_set:
        bluecat.deploy_server_services(server, 'DHCP')
    monitor_server_deployment(server_set)


def monitor_server_deployment(server_set):
    for server in server_set:
        status = bluecat.get_deployment_status(server)
        logger.info('{} - {}'.format(DATA.ADONIS_ID_MAP[server], DATA.DEPLOYMENT_STATUS[status]))
    while len(server_set) > 0:
        sleep(2)
        servers = list(server_set)
        for server in servers:
            status = bluecat.get_deployment_status(server)
            if status not in [-1, 0, 1]:
                logger.info('{} - {}'.format(DATA.ADONIS_ID_MAP[server], DATA.DEPLOYMENT_STATUS[status]))
                server_set.remove(server)


def create_dns_a_record(name, ip):
    job_id = creds['rundeck']['jobs']['dns-a']
    url = 'https://rundeck.spectrum-health.org:4443/api/21/job/{}/run'.format(job_id)
    apikey = creds['rundeck']['apikey']
    params = {
        'format': 'json',
        'authtoken': apikey
    }
    body = {
        'argString': '-IPAddress {} -RecordToAdd {}'.format(ip, name)
    }
    response = requests.post(url, json=body, params=params)
    sleep(1)
    return response


def delete_dns_a_record(name):
    job_id = creds['rundeck']['jobs']['dns-a-remove']
    url = 'https://rundeck.spectrum-health.org:4443/api/21/job/{}/run'.format(job_id)
    apikey = creds['rundeck']['apikey']
    params = {
        'format': 'json',
        'authtoken': apikey
    }
    body = {
        'argString': '-RecordToDelete {}'.format(name)
    }
    response = requests.post(url, json=body, params=params)
    sleep(1)
    return response


def add_single_operation_args(parser):
    parser.add_argument('hostname')
    me_group = parser.add_mutually_exclusive_group(required=True)
    me_group.add_argument('-n', '--network', help='network address within desired subnet')
    me_group.add_argument('-e', '--environment', help='environment to assign IP from')
    parser.add_argument('-c', '--creds', help='filepath to read in json credentials')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    parser.add_argument('--nowait', action='store_true', help='do NOT wait for deploy before printing results')


def add_dhcp_single_operation_args(parser):
    add_single_operation_args(parser)
    parser.add_argument('mac')


def add_bulk_operation_args(parser):
    parser.add_argument('filepath', help='path to csv file containing record info')
    parser.add_argument('-c', '--creds', help='filepath to read in json credentials')
    me_group= parser.add_mutually_exclusive_group()
    me_group.add_argument('--create', action='store_true', help='create all records in csv, default action')
    me_group.add_argument('--delete', action='store_true', help='delete all records in csv')
    me_group.add_argument('--update', action='store_true', help='update all records in csv')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')


def main():
    # Main Argument Parser "pybluecat"
    parser = argparse.ArgumentParser(prog='Bluecat BAM CLI Tool')
    subparsers = parser.add_subparsers(title='Subcommands', help='subparsers command help')

    ## "pybluecat" sub-parser: "static"
    parser_static = subparsers.add_parser('static', help='static IP record manipulation')
    sub_static = parser_static.add_subparsers(title='Subcommands', help='options for static records')
    ## "pybluecat" sub-parser: "dhcp"
    parser_dhcp = subparsers.add_parser('dhcp', help='dhcp IP record manipulation')
    sub_dhcp = parser_dhcp.add_subparsers(title='Subcommands', help='options for dhcp records')
    ## "pybluecat" sub-parser: "search"
    parser_search = subparsers.add_parser('search', help='search BAM for Objects')
    sub_search = parser_search.add_subparsers(title='Subcommands', help='options for dhcp records')

    ### pybluecat static create
    parser_static_create = sub_static.add_parser('create', help='create a static IP reservation')
    add_single_operation_args(parser_static_create)

    ### pybluecat static delete
    parser_static_delete = sub_static.add_parser('delete', help='delete a static IP reservation')
    add_single_operation_args(parser_static_delete)

    ### pybluecat static update
    parser_static_update = sub_static.add_parser('update', help='update a static IP reservation')
    add_single_operation_args(parser_static_update)

    ### pybluecat static bulk
    parser_static_bulk = sub_static.add_parser('bulk', help='create bulk static records from csv')
    add_bulk_operation_args(parser_static_bulk)

    ### pybluecat dhcp create
    parser_dhcp_create = sub_dhcp.add_parser('create', help='create a dhcp IP reservation')
    add_dhcp_single_operation_args(parser_dhcp_create)
    parser_dhcp_create.set_defaults(func=create_dhcp)

    ### pybluecat dhcp delete
    parser_dhcp_delete = sub_dhcp.add_parser('delete', help='delete a dhcp IP reservation')
    add_dhcp_single_operation_args(parser_dhcp_delete)

    ### pybluecat dhcp update
    parser_dhcp_update = sub_dhcp.add_parser('update', help='update a dhcp IP reservation')
    add_dhcp_single_operation_args(parser_dhcp_update)

    ### pybluecat dhcp bulk
    parser_dhcp_bulk = sub_dhcp.add_parser('bulk', help='create bulk dhcp records from csv')
    add_bulk_operation_args(parser_dhcp_bulk)
    parser_dhcp_bulk.set_defaults(func=handle_dhcp_bulk)

    # Parse the args from any and all parsers
    args = parser.parse_args()

    # Setup BAM-CLI logger
    global logger
    logger = logging.getLogger('pybluecat-cli')
    loglevel = args.loglevel.upper()
    if loglevel in ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']:
        level = getattr(logging, loglevel)
        logger.setLevel(level=level)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)

    # Get credentials and client
    global creds
    creds = get_creds(args)
    global bluecat
    bluecat = get_client(loglevel=loglevel)

    # Run the relevant function
    output = args.func(args)
    print(json.dumps(output, indent=2, sort_keys=True))


if __name__ == '__main__':
    main()



#!/usr/bin/python

from proteus import Proteus

hostname = 'Forrest-Test'
mac_addr = 'deadbeef0987'

p = Proteus()
#p.getIP4Network('10.168.161.1')
#ip_addr =  p.getNextIP4Address()
#p.assignIP4Address(hostname, ip_addr)
#p.queueServers()
#print p.network
#print p.servers_to_deploy
#p.deployServerConfig()
#p.monitorServerDeployment()
#p.logout()

e = p.getMACAddress(mac_addr)
r = p.getLinkedEntities(e)
for i in r:
  print i

p.logout()



import json
import os
from pybluecat.bam import BAM
from pybluecat.exceptions import *


def get_creds(filepath=None):
    if filepath is None:
        os_name = os.name
        if os_name == 'posix':
            filepath = os.environ['HOME'] + '/.bluecat'
        elif os_name == 'nt':
            filepath = os.environ['HOMEPATH'] + '\\.bluecat'
        else:
            filepath = '.bluecat'
    with open(filepath) as f:
        creds = json.load(f)
    if 'bluecat' in creds:
        creds = creds['bluecat']
    return creds

def prop_s2d(propString):
    if propString is None:
        return None
    else:
        return {p[0]: p[1] for p in [pair.split('=') for pair in propString.split('|')[:-1]]}

def prop_d2s(propDict):
    if propDict is None:
        return None
    else:
        return '|'.join(['='.join(pair) for pair in propDict.items()]) + '|'

def entity_to_json(entity):
    entity['properties'] = prop_s2d(entity['properties'])
    return entity

def json_to_entity(self, entity):
    entity['properties'] = prop_d2s(entity['properties'])
    return entity




#!/usr/bin/python
import argparse
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def print_ip_info(ip_obj, network):
    print('Hostname   : {}'.format(ip_obj['name']))
    print('MAC Address: {}'.format(ip_obj['properties']['macAddress']))
    print('IP Address: {}'.format(ip_obj['properties']['address']))
    print('Subnet Mask: {}'.format(str(network.netmask)))
    print('Gateway    : {}'.format(str(network.network_address + 1)))
    for key in ip_obj['properties']:
        pass


def search_mac(session, mac):
    formatted_mac = mac.replace('.', '').replace(':', '').replace('-', '')
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('hostname', help='hostname')
    parser.add_argument('mac_addr', help='MAC Address')
    parser.add_argument('ip_addr', help='IP desired for reservation, also defines the network if "first_available" flag is set')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    parser.add_argument('--location', default='', help='Location of Device')
    parser.add_argument('--notes', default='', help='Additional notes about the reservation')
    parser.add_argument('--owner', default='', help='Owner of the device')
    parser.add_argument('--request_num', default='', help='ServiceDesk Request Number')
    parser.add_argument('--first_available', action='store_true', help='Use first available IP as fallback option')
    parser.add_argument('--no_wait', action='store_true', help='do NOT wait for deployment')
    parser.add_argument('--no_deploy', action='store_true', help='Do NOT deploy saved changes')
    args = parser.parse_args()

    # Setup console logging
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(level=loglevel)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.propagate = False

    # Setup args and vars
    hostname = args.hostname
    mac = args.mac_addr.replace('.', '').replace(':', '').replace('-', '')
    ip = ip_address(unicode(args.ip_addr))
    action = 'MAKE_DHCP_RESERVED'
    creds = pybluecat.get_creds(args.creds)
    properties = {
        'Location': args.location,
        'Notes': args.notes,
        'Owner': args.owner,
        'RequestNum': args.request_num
    }
    output_object = {}

    # Initialize flow-control boolean vars
    deploy_needed = False
    ip_is_eligible = False
    mac_already_reserved = False
    desired_ip_reserved = False
    error_message = ''

    # Create instance using 'with' so cleanup of session is automatic
    with pybluecat.BAM(**creds) as bam:
        # Determine the Network of the IP Address
        try:
            logger.info('Getting Network info from ip: {}'.format(str(ip)))
            net_entity = bam.get_network(str(ip))
            net_obj = pybluecat.entity_to_json(net_entity)
            network = ip_network(unicode(net_obj['properties']['CIDR']))
            if network.prefixlen <= 24:
                dhcp_offset_ip = str(network.network_address + 31)
            else:
                dhcp_offset_ip = None
            logger.debug(json.dumps(net_obj, indent=2))
        except pybluecat.exceptions.BluecatError as e:
            logger.error('Could not determine the target network or network did not exist')
            exit(str(e))

        # Determine if IP eligible to be reserved, not Broadcast, Network, Gateway, etc...
        if ip == network.network_address:
            error_message = 'Cannot reserve the Network Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip == network.broadcast_address:
            error_message = 'Cannot reserve the Broadcast Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip == network.network_address + 1:
            error_message = 'Cannot reserve the Gateway Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip in [network.network_address + 2, network.network_address + 3]:
            error_message = 'Cannot reserve a GSLB/VRRP Address: {}'.format(str(ip))
            logger.info(error_message)
        elif network.prefixlen <= 24 and ip in [network.network_address + i for i in xrange(4, 31)]:
            error_message = 'Cannot reserve Address in Telecom-Reserved Space: {}'.format(str(ip))
            logger.info(error_message)
        else:
            ip_is_eligible = True

        # Determine if MAC is already associated with a reservation in target network
        logger.info('Searching for existing reservations for MAC: {}'.format(mac))
        try:
            linked_entities = search_mac(bam, mac)
            for entity in linked_entities:
                if entity['type'] == 'IP4Address':
                    ip_obj = pybluecat.entity_to_json(entity)
                    logger.debug(json.dumps(ip_obj, indent=2, sort_keys=True))
                    ip_obj_address = ip_address(unicode(ip_obj['properties']['address']))
                    # Break loop as soon as match is found, return the matched reservation
                    if ip_obj_address in network and ip_obj['properties']['state'] == 'DHCP_RESERVED':
                        logger.info('Found existing reservation in target network')
                        mac_already_reserved = True
                        output_object = {
                            'status': 'Found existing reservation',
                            'message': 'MAC Address has existing reservation in current network.',
                            'reservation': ip_obj
                        }
                        break
        # If MAC not found handle the error, otherwise exit script with error code 1
        except pybluecat.exceptions.BluecatError as e:
            if 'Object was not found' in e.message:
                logger.info('MAC address does not currently exist in Bluecat')
            else:
                logger.error('Failed searching for MAC')
                exit(str(e))

        # If we didn't find an existing reservation, try to get a new one
        if not mac_already_reserved:
            logger.info('MAC address has no current reservations in target network')
            # Try to create the Reservation using the literal IP given as arg
            if ip_is_eligible:
                logger.info('Sending reservation request to Bluecat for IP: {}'.format(str(ip)))
                try:
                    ip_id = bam.assign_ip_address(hostname, str(ip), mac, action, properties)
                    ip_entity = bam.get_entity_by_id(ip_id)
                    ip_obj = pybluecat.entity_to_json(ip_entity)
                    desired_ip_reserved = True
                    deploy_needed = True
                    output_object = {
                        'status': 'New Reservation Create',
                        'message': 'Desired IP Address was Available and Assigned',
                        'reservation': ip_obj
                    }
                # Handle Bluecat errors, shouldn't see anything other than Dupes at this point
                except pybluecat.exceptions.BluecatError as e:
                    if 'Duplicate' in e.message:
                        error_message = 'Desired IP is already in Use'
                        logger.info(error_message)
                        ip_entity = bam.get_ip_address(str(ip))
                        conflicting_ip = pybluecat.entity_to_json(ip_entity)
                    elif 'already used by another IP within the same network' in e.message:
                        # Should've been caught earlier
                        error_message = e.message
                        logger.error(error_message)
                    else:
                        # Probably a system failure :(
                        logger.error(e.message)

            # If we were not able to reserve the desired IP for whatever reason...
            if not desired_ip_reserved:
                # Get next available IP if flag is set
                if args.first_available:
                    try:
                        logger.info('Sending request to Bluecat to assign next available address in network: {}'.format(str(network)))
                        ip_entity = bam.assign_next_ip_address(net_obj['id'], hostname, mac, action, properties, dhcp_offset_ip)
                        ip_obj = pybluecat.entity_to_json(ip_entity)
                        logger.debug(json.dumps(ip_obj, indent=2, sort_keys=True))
                        output_object = {
                            'status': 'New Reservation Created',
                            'message': [
                                error_message,
                                'First Available IP Address was assigned'
                            ],
                            'reservation': ip_obj
                        }
                        deploy_needed = True
                    # Expected errors here would be network out of addresses
                    except pybluecat.exceptions.BluecatError as e:
                        output_object = {
                            'status': 'Reservation could not be completed',
                            'message': [
                                error_message,
                                e.message
                            ]
                        }
                        logger.error('Unable to assign next available IP address')
                # Should hit this only when target IP in ineligible and __first_available is not set
                else:
                    output_object = {
                        'status': 'Reservation could not be completed',
                        'message': [
                            error_message,
                            'Try using the --first_available flag or choose a different IP Address'
                        ],
                        'conflicting_ip': conflicting_ip
                    }

        # Deploy DHCP changes if no_deploy flag isn't set
        if deploy_needed and not args.no_deploy:
            logger.info('Deploying Config changes')
            server_set = bam.queue_servers(net_obj['id'])
            for server in server_set:
                logger.info('Deploying changes to server: {}'.format(str(server)))
                bam.deploy_server_services(server, 'DHCP')
            # Monitor the deployment through completion unless no_wait flag is set
            if not args.no_wait:
                logger.info('Waiting for deployment to finish')
                bam.monitor_server_deployment(server_set)
                logger.info('Deployment finished')
            else:
                logger.info('Monitoring of deployment is being skipped')

    # Finally output the IP info
    #print_ip_info(ip_obj)
    print(json.dumps(output_object, indent=2, sort_keys=True))


if __name__ == '__main__':
    main()




#!/usr/bin/python
import argparse
import csv
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def print_ip_info(ip_obj, network):
    print('Hostname   : {}'.format(ip_obj['name']))
    print('MAC Address: {}'.format(ip_obj['properties']['macAddress']))
    print('IP Address: {}'.format(ip_obj['properties']['address']))
    print('Subnet Mask: {}'.format(str(network.netmask)))
    print('Gateway    : {}'.format(str(network.network_address + 1)))
    for key in ip_obj['properties']:
        pass


def search_mac(session, mac):
    formatted_mac = mac.replace('.', '').replace(':', '').replace('-', '')
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results


def is_mac_in_use(mac, network, bam, logger):
    output_object = None
    # Determine if MAC is already associated with a reservation in target network
    logger.info('Searching for existing reservations for MAC: {}'.format(mac))
    try:
        linked_entities = search_mac(bam, mac)
        for entity in linked_entities:
            logging.debug('Entity: {}'.format(json.dumps(entity, indent=2, sort_keys=True)))
            if entity['type'] == 'IP4Address':
                ip_obj = pybluecat.entity_to_json(entity)
                logger.debug(json.dumps(ip_obj, indent=2, sort_keys=True))
                ip_obj_address = ip_address(unicode(ip_obj['properties']['address']))
                # Break loop as soon as match is found, return the matched reservation
                if ip_obj_address in network and ip_obj['properties']['state'] == 'DHCP_RESERVED':
                    logger.info('Found existing reservation in target network')
                    output_object = {
                        'status': 'Found existing reservation',
                        'message': 'MAC Address has existing reservation in current network.',
                        'reservation': ip_obj
                    }
                    break
    # If MAC not found handle the error, otherwise exit script with error code 1
    except pybluecat.exceptions.BluecatError as e:
        if 'Object was not found' in e.message:
            logger.info('MAC address does not currently exist in Bluecat')
        else:
            logger.error('Failed searching for MAC')
            exit(str(e))
    return output_object


def is_ip_eligible(ip, network, logger):
    # Determine if IP eligible to be reserved, not Broadcast, Network, Gateway, etc...
    if ip == network.network_address:
        error_message = 'Cannot reserve the Network Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip == network.broadcast_address:
        error_message = 'Cannot reserve the Broadcast Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip == network.network_address + 1:
        error_message = 'Cannot reserve the Gateway Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip in [network.network_address + 2, network.network_address + 3]:
        error_message = 'Cannot reserve a GSLB/VRRP Address: {}'.format(str(ip))
        logger.info(error_message)
    elif network.prefixlen <= 24 and ip in [network.network_address + i for i in xrange(4, 31)]:
        error_message = 'Cannot reserve Address in Telecom-Reserved Space: {}'.format(str(ip))
        logger.info(error_message)
    else:
        error_message = None
    return error_message


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('csvpath', help='path to csv file')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    parser.add_argument('--first_available', action='store_true', help='Use first available IP as fallback option')
    parser.add_argument('--no_wait', action='store_true', help='do NOT wait for deployment')
    parser.add_argument('--no_deploy', action='store_true', help='Do NOT deploy saved changes')
    args = parser.parse_args()

    # Setup console logging
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(level=loglevel)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.propagate = False

    # Setup args and vars
    with open(args.csvpath) as f:
        # fieldnames = ['hostname', 'macaddress', 'ipaddress', 'location', 'owner', 'notes', 'requestnum']
        reader = csv.DictReader(f)  # , fieldnames)
        reservations = [row for row in reader]
        logger.debug(json.dumps(reservations, indent=2, sort_keys=True))

    # Initialize flow-control boolean vars
    deploy_needed = False
    action = 'MAKE_DHCP_RESERVED'
    creds = pybluecat.get_creds(args.creds)

    # Create instance using 'with' so cleanup of session is automatic
    output_list = []
    server_set = set()
    with pybluecat.BAM(**creds) as bam:
        for reservation in reservations:
            ip_is_eligible = False
            mac_already_reserved = False
            desired_ip_reserved = False
            hostname = reservation['hostname']
            mac = reservation['macaddress']
            ip = ip_address(unicode(reservation['ipaddress']))
            properties = {
                'Location': reservation['location'],
                'Notes': reservation['notes'],
                'Owner': reservation['owner'],
                'RequestNum': reservation['requestnum']
            }
            logger.info('Processing Row: Name={}, MAC={}, IP={}'.format(hostname, mac, ip))
            output_object = {}
            error_message = ''

            # Determine the Network of the IP Address
            try:
                logger.info('Getting Network info from ip: {}'.format(str(ip)))
                net_entity = bam.get_network(str(ip))
                net_obj = pybluecat.entity_to_json(net_entity)
                network = ip_network(unicode(net_obj['properties']['CIDR']))
                if network.prefixlen <= 24:
                    dhcp_offset_ip = str(network.network_address + 31)
                else:
                    dhcp_offset_ip = None
                logger.debug(json.dumps(net_obj, indent=2))
            except pybluecat.exceptions.BluecatError as e:
                logger.error('Could not determine the target network or network did not exist')
                exit(str(e))

            # Determine if IP eligible to be reserved, not Broadcast, Network, Gateway, etc...
            error_message = is_ip_eligible(ip, network, logger)
            if error_message is None:
                ip_is_eligible = True

            # Determine if MAC is already associated with a reservation in target network
            mac_res = is_mac_in_use(mac, network, bam, logger)
            if mac_res is not None:
                mac_already_reserved = True
                output_object = mac_res

            # If we didn't find an existing reservation, try to get a new one
            if not mac_already_reserved:
                logger.info('MAC address has no current reservations in target network')
                # Try to create the Reservation using the literal IP given as arg
                if ip_is_eligible:
                    logger.info('Sending reservation request to Bluecat for IP: {}'.format(str(ip)))
                    try:
                        ip_id = bam.assign_ip_address(hostname, str(ip), mac, action, properties)
                        ip_entity = bam.get_entity_by_id(ip_id)
                        ip_obj = pybluecat.entity_to_json(ip_entity)
                        logger.info('Reservation for {} was successful'.format(ip_obj['properties']['address']))
                        logger.debug(json.dumps(ip_obj, indent=2, sort_keys=True))
                        desired_ip_reserved = True
                        deploy_needed = True
                        output_object = {
                            'status': 'New Reservation Create',
                            'message': 'Desired IP Address was Available and Assigned',
                            'reservation': ip_obj
                        }
                    # Handle Bluecat errors, shouldn't see anything other than Dupes at this point
                    except pybluecat.exceptions.BluecatError as e:
                        if 'Duplicate' in e.message:
                            error_message = 'Desired IP is already in Use'
                            logger.info(error_message)
                            ip_entity = bam.get_ip_address(str(ip))
                            conflicting_ip = pybluecat.entity_to_json(ip_entity)
                        elif 'already used by another IP within the same network' in e.message:
                            # Should've been caught earlier
                            error_message = e.message
                            logger.error(error_message)
                        else:
                            # Probably a system failure :(
                            logger.error(e.message)

                # If we were not able to reserve the desired IP for whatever reason...
                if not desired_ip_reserved:
                    # Get next available IP if flag is set
                    if args.first_available:
                        try:
                            logger.info('Sending request to Bluecat to assign next available address in network: {}'.format(str(network)))
                            ip_entity = bam.assign_next_ip_address(net_obj['id'], hostname, mac, action, properties, dhcp_offset_ip)
                            ip_obj = pybluecat.entity_to_json(ip_entity)
                            logger.info('Reservation for {} was successful'.format(ip_obj['properties']['address']))
                            logger.debug(json.dumps(ip_obj, indent=2, sort_keys=True))
                            output_object = {
                                'status': 'New Reservation Created',
                                'message': [
                                    error_message,
                                    'First Available IP Address was assigned'
                                ],
                                'reservation': ip_obj
                            }
                            deploy_needed = True
                        # Expected errors here would be network out of addresses
                        except pybluecat.exceptions.BluecatError as e:
                            output_object = {
                                'status': 'Reservation could not be completed',
                                'message': [
                                    error_message,
                                    e.message
                                ]
                            }
                            logger.error('Unable to assign next available IP address')
                    # Should hit this only when target IP in ineligible and __first_available is not set
                    else:
                        output_object = {
                            'status': 'Reservation could not be completed',
                            'message': [
                                error_message,
                                'Try using the --first_available flag or choose a different IP Address'
                            ],
                            'conflicting_ip': conflicting_ip
                        }
            output_list.append(output_object)
            if deploy_needed and not args.no_deploy:
                logger.info('Adding Bluecat servers to deploy queue')
                server_set = bam.queue_servers(net_obj['id'], server_set)

        # Deploy DHCP changes if no_deploy flag isn't set
        if deploy_needed and not args.no_deploy:
            logger.info('Deploying Config changes')
            for server in server_set:
                logger.info('Deploying changes to server: {}'.format(str(server)))
                bam.deploy_server_services(server, 'DHCP')
            # Monitor the deployment through completion unless no_wait flag is set
            if not args.no_wait:
                logger.info('Waiting for deployment to finish')
                bam.monitor_server_deployment(server_set)
                logger.info('Deployment finished')
            else:
                logger.info('Monitoring of deployment is being skipped')

        # Finally output the IP info
        #print_ip_info(ip_obj)
        print(json.dumps(output_list, indent=2, sort_keys=True))


if __name__ == '__main__':
    main()




#!/usr/bin/python
import argparse
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def format_mac(mac):
    return mac.replace('.', '').replace(':', '').replace('-', '').lower()


def search_mac(session, mac):
    formatted_mac = format_mac(mac)
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results


def is_ip_off_limits(ip, network):
    off_limits = True
    error_message = ''
    # Determine if IP eligible to be reserved, not Broadcast, Network, Gateway, etc...
    if ip == network.network_address:
        error_message = 'Cannot Delete the Network Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip == network.broadcast_address:
        error_message = 'Cannot Delete the Broadcast Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip == network.network_address + 1:
        error_message = 'Cannot Delete the Gateway Address: {}'.format(str(ip))
        logger.info(error_message)
    elif ip in [network.network_address + 2, network.network_address + 3]:
        error_message = 'Cannot Delete a HSRP/VRRP Address: {}'.format(str(ip))
        logger.info(error_message)
    elif network.prefixlen <= 24 and ip in [network.network_address + i for i in xrange(4, 31)]:
        error_message = 'Cannot Delete Address in Telecom-Reserved Space: {}'.format(str(ip))
        logger.info(error_message)
    else:
        off_limits = False
    return off_limits, error_message


def main():
    parser = argparse.ArgumentParser()
    field_group = parser.add_argument_group('Matchable Fields', 'Provide at least one of the following fields to delete a reservation')
    parser.add_argument('--hostname', help='hostname of reservation(s) to delete')
    parser.add_argument('--mac_addr', help='MAC Address of reservation(s) to delete')
    parser.add_argument('--ip_addr', help='IP Address of reservation to delete')
    match_group = parser.add_mutually_exclusive_group(required=True)
    match_group.add_argument('--match_all', action='store_true', help='Delete only objects that match ALL conditions')
    match_group.add_argument('--match_any', action='store_true', help='Delete any object that matches ANY condition')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    parser.add_argument('--no_wait', action='store_true', help='do NOT wait for deployment')
    parser.add_argument('--no_deploy', action='store_true', help='Do NOT deploy saved changes')
    args = parser.parse_args()

    # Setup console logging
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(level=loglevel)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.propagate = False

    # Setup args and vars
    # None values are passed along to be ignored later
    # Rundeck passes in empty strings instead, so translate those to None
    if args.hostname is not None:
        if args.hostname.strip() == '':
            hostname = None
        else:
            hostname = args.hostname
    else:
        hostname = None
    if args.mac_addr is not None:
        if args.mac_addr.strip() == '':
            mac = None
        else:
            mac = format_mac(args.mac_addr)
    else:
        mac = None
    if args.ip_addr is not None:
        if args.ip_addr.strip() == '':
            ip = None
        else:
            ip = ip_address(unicode(args.ip_addr))
    else:
        ip = None
    creds = pybluecat.get_creds(args.creds)
    output_object = {
        'status': '',
        'message': [],
        'deleted_reservations': []
    }

    # Exit Immediately if no fields were provided
    if all(field is None for field in [hostname, mac, ip]):
        output_object['status'] = 'No Reservations were Deleted'
        output_object['message'].append('No Fields were populated')
        output_object['message'].append('Please provided at least a Hostname, MAC_Address, or IP_Address')
        print(json.dumps(output_object, indent=2, sort_keys=True))
        exit()

    # Initialize flow-control boolean vars
    deploy_needed = False
    ip_is_eligible = False
    error_message = ''
    ips_seen_so_far = []
    candidate_reservations = []
    reservations_to_delete = []
    deleted_reservations = []

    # Create instance using 'with' so cleanup of session is automatic
    with pybluecat.BAM(**creds) as bam:
        # Get reservation matching the hostname, if given
        if hostname is not None:
            logger.info('Hostname provided ({}), searching for reservations'.format(hostname))
            results = bam.search_ip_by_name(hostname)
            logger.debug(json.dumps(results, indent=2, sort_keys=True))
            candidate_reservations += results
        # Get reservations matching the IP Address, if given
        if ip is not None:
            logger.info('IP Address provided ({}), searching for reservations'.format(str(ip)))
            result = bam.get_ip_address(str(ip))
            logger.debug(json.dumps(result, indent=2, sort_keys=True))
            candidate_reservations.append(result)
        # Get reservations matching the MAC Address, if given
        if mac is not None:
            logger.info('MAC Address provided ({}), searching for reservations'.format(args.mac_addr))
            results = search_mac(bam, mac)
            logger.debug(json.dumps(results, indent=2, sort_keys=True))
            candidate_reservations += results

        # Determine which of the matched reservations qualify for deletion
        logger.info('Checking all gathered reservations for deletion eligibility')
        for reservation in candidate_reservations:
            reservation = pybluecat.entity_to_json(reservation)
            # Skip NULL objects, can happen when MAC searches turn up nothing
            if reservation['properties'] is None:
                logger.info('Ignoring NULL object')
                continue
            # Gather details about the ip_object
            res_name = reservation['name']
            res_ip = reservation['properties']['address']
            if 'macAddress' in reservation['properties']:
                res_mac = reservation['properties']['macAddress']
            else:
                res_mac = ''
            logger.info('NAME: {}, IP: {}, MAC: {}'.format(res_name, res_ip, res_mac))
            # If IP Object state is not "DHCP_RESERVED" DO NOT DELETE
            if reservation['properties']['state'] != 'DHCP_RESERVED':
                logger.info('Reservation ineligible for deletion, ignoring objects with state: {}'.format(reservation['properties']['state']))
                continue
            # Ignore duplicate objects from the multiple searches
            if res_ip in ips_seen_so_far:
                logger.info('Duplicate entry for IP {}, skipping'.format(res_ip))
                continue
            # If --match_all is set, ensure ALL set fields match the reservation
            if args.match_all:
                # If --hostname is set and doesn't match, skip to next reservation
                if hostname is not None and res_name.lower() != hostname.lower():
                    logger.info('Reservation doesn\'t match --hostname filter with --match_all set, ineligible for deletion')
                    continue
                # If --ip_addr is set and doesn't match, skip to next reservation
                if ip is not None and res_ip != str(ip):
                    logger.info('Reservation doesn\'t match --ip_addr filter with --match_all set, ineligible for deletion')
                    continue
                # If --mac_addr is set and doesn't match, skip to next reservation
                if mac is not None and format_mac(res_mac) != mac:
                    logger.info('Reservation doesn\'t match --mac_addr filter with --match_all set, ineligible for deletion')
                    continue
            # Determine the Network of the IP Address
            try:
                logger.info('Getting Network info from ip: {}'.format(res_ip))
                net_entity = bam.get_network(res_ip)
                net_obj = pybluecat.entity_to_json(net_entity)
                network = ip_network(unicode(net_obj['properties']['CIDR']))
                logger.debug(json.dumps(net_obj, indent=2))
            except pybluecat.exceptions.BluecatError as e:
                # This would be really odd... But try to continue with the other reservations
                logger.error('Could not determine the target network or network did not exist')
                logger.error(e.message)
                continue
            # Continue on with either all reservations (if --match_any) or whatever is left from the --match_all block
            off_limits, error_message = is_ip_off_limits(res_ip, network)
            if off_limits:
                logger.info('Reservation ineligible for deletion: {}'.format(error_message))
                continue
            # Finally, mark the reservation for deletion
            reservations_to_delete.append((reservation, net_obj))
            ips_seen_so_far.append(res_ip)

        # Start deleting!
        for reservation, net_obj in reservations_to_delete:
            logger.info('Deleting DHCP Reservation: {}'.format(json.dumps(reservation, indent=2, sort_keys=True)))
            response = bam.delete(reservation['id'])
            output_object['deleted_reservations'].append(reservation)
            output_object['message'].append('DELETED: {}, {}'.format(reservation['properties']['address'], reservation['name']))
            deleted_reservations.append((reservation, net_obj))
        if len(deleted_reservations) > 0:
            deploy_needed = True
            output_object['status'] = 'Reservations have been deleted, config NOT deployed'
        else:
            output_object['status'] = 'No Reservations were Deleted'

        server_set = set()
        # Deploy DHCP changes if no_deploy flag isn't set
        if deploy_needed and not args.no_deploy:
            # Build unique set of DHCP servers
            logger.info('Gathering list of BAM servers to push updated configs to')
            for reservation, net_obj in deleted_reservations:
                server_set = server_set | bam.queue_servers(net_obj['id'])
            # Deploy changes to all servers in set
            logger.info('Deploying Config changes')
            for server in server_set:
                logger.info('Deploying changes to server: {}'.format(str(server)))
                bam.deploy_server_services(server, 'DHCP')
                output_object['message'].append('Server {} queued for deployment'.format(str(server)))
            # Monitor the deployment through completion unless no_wait flag is set
            if not args.no_wait:
                logger.info('Waiting for deployment to finish')
                bam.monitor_server_deployment(server_set)
                output_object['status'] = 'Reservations have been deleted, configs have been deployed'
                logger.info('Deployment finished')
            else:
                output_object['status'] = 'Reservations have been deleted, configs queued for deployment'
                logger.info('Monitoring of deployment is being skipped')

    # Finally output the IP info
    print(json.dumps(output_object, indent=2, sort_keys=True))


if __name__ == '__main__':
    main()




#!/usr/bin/python
import requests
import json
import logging
from ipaddress import ip_address, ip_network
from pybluecat import data as DATA
from pybluecat.data import *
from pybluecat.exceptions import BluecatError
from time import sleep


class BAM:
    """About the Bluecat REST API:
    REST APIs have many similarities with the widely used SOAP-based APIs supported by Address
    Manager. However, there are a few differences between REST interface and existing SOAP
    implementation:
    - The names of API methods in REST remain the same as that of SOAP APIs.
    - Signatures of all methods including input and output parameters in REST are the same as in SOAP.
    - In REST API, various primitive request parameters such as int, long and String are expected as URL
    query parameters. Whereas in SOAP, all the request parameters are communicated as part of XML
    body.
    - Complex parameter types such as APIEntity or APIDeploymentOption need to be passed as a part of
    HTTP body of the call in JSON format.
    """

    # Constants
    PROXIES = {'http': None, 'https': None}
    IP_ACTION_VALUES = [
        'MAKE_STATIC',
        'MAKE_RESERVED',
        'MAKE_DHCP_RESERVED'
    ]

    def __init__(self, hostname=None, username=None, password=None, configName='Spectrum Health', loglevel='CRITICAL'):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.history = []
        self.lastCall = None
        self.loglevel = loglevel
        self.logger = self.set_loglevel('pybluecat', loglevel)
        self.py_logger = self.set_loglevel('py.warnings', loglevel)
        self.session = self.init_session()
        self.baseUrl = 'https://{h}/Services/REST/v1/'.format(h=hostname)
        if all(param is not None for param in [hostname, username, password]):
            self.login(hostname, username, password)
            self.config = self.getConfig()
        else:
            self.config = None

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.logout()

    def init_session(self, proxies={'http': None, 'https': None},
                     headers={'Content-Type': 'application/json'},
                     ssl_verify=False):
        session = requests.Session()
        session.proxies.update(proxies)
        session.headers.update(headers)
        session.verify = ssl_verify
        if not ssl_verify:
            logging.captureWarnings(True)
        return session

    def set_loglevel(self, logger_name, loglevel):
        logger = logging.getLogger(logger_name)
        loglevel = loglevel.upper()
        if loglevel in ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']:
            level = getattr(logging, loglevel)
            # logging.basicConfig(level=level)
            logger.setLevel(level=level)
        console_handler = logging.StreamHandler()
        logger.addHandler(console_handler)
        return logger

    ################################################################
    # DECORATORS
    ################################################################

    def rest_call(httpMethod):
        def outer(func):
            def inner(self, *args, **kwargs):
                method, params, data = func(self, *args, **kwargs)
                url = self.baseUrl + method
                methodMap = {
                    'delete': self.session.delete,
                    'get': self.session.get,
                    'post': self.session.post,
                    'put': self.session.put
                }
                response = methodMap[httpMethod](url, params=params, json=data)
                self.logger.debug('Request URL: {}'.format(response.request.url))
                self.logger.debug('Response Code: {}'.format(response.status_code))
                self.lastCall = response
                self.history.append(response)
                # Handle non-200 responses
                if response.status_code != 200:
                    raise BluecatError(response)
                try:
                    data = response.json()
                    self.logger.debug('Response Body: {}'.format(json.dumps(data, indent=2, sort_keys=True)))
                except Exception:
                    data = response.content
                    self.logger.debug('Response Body: {}'.format(data))
                return data
            return inner
        return outer

    ################################################################
    # HELPERS
    ################################################################

    def prop_s2d(self, propString):
        if propString is None:
            return None
        else:
            return {p[0]: p[1] for p in [pair.split('=') for pair in propString.split('|')[:-1]]}

    def prop_d2s(self, propDict):
        if propDict is None:
            return None
        else:
            return '|'.join(['='.join(pair) for pair in propDict.items()]) + '|'

    def entity_to_json(self, entity):
        entity['properties'] = self.prop_s2d(entity['properties'])
        return entity

    def json_to_entity(self, entity):
        entity['properties'] = self.prop_d2s(entity['properties'])
        return entity

    ################################################################
    # GENERAL STUFF
    ################################################################

    def login(self, host, username, password):
        method = 'login'
        params = {
            'username': username,
            'password': password
        }
        try:
            response = self.session.get(self.baseUrl + method, params=params)
            self.logger.info(response.content)
            # authToken = response.text.split('BAMAuthToken: ')[1].split(' <- ')[0]
            authToken = response.text.split('-> ')[1].split(' <-')[0]
            self.session.headers.update({'Authorization': str(authToken)})
        except:
            self.logger.error('ERROR: Login Failed')
        return response

    @rest_call('get')
    def logout(self):
        method = 'logout'
        params = None
        data = None
        return method, params, data

    @rest_call('get')
    def get_entity_by_name(self, parentId, name, objType):
        method = 'getEntityByName'
        params = {
            'parentId': parentId,
            'name': name,
            'type': objType
        }
        data = None
        return method, params, data

    def getConfig(self):
        return self.get_entity_by_name(0, 'Spectrum Health', 'Configuration')

    @rest_call('get')
    def get_entities(self, parent_id, obj_type, start=0, count=1000):
        method = 'getEntities'
        params = {
            'parentId': parent_id,
            'type': obj_type,
            'start': start,
            'count': count
        }
        data = None
        return method, params, data

    def get_networks(self, parent_id, start=0, count=1000):
        return self.get_entities(parent_id, 'IP4Network', start, count)

    @rest_call('get')
    def get_entity_by_id(self, entityId):
        method = 'getEntityById'
        params = {
            'id': entityId
        }
        data = None
        return method, params, data

    @rest_call('get')
    def get_linked_entities(self, entityId, linkedType='IP4Address', start=0, count=100):
        method = 'getLinkedEntities'
        params = {
            'entityId': entityId,
            'type': linkedType,
            'start': start,
            'count': count
        }
        data = None
        return method, params, data

    @rest_call('delete')
    def delete(self, entity_id):
        method = 'delete'
        params = {
            'objectId': entity_id
        }
        data = None
        return method, params, data

    @rest_call('put')
    def update(self, entity):
        method = 'update'
        params = None
        data = entity
        return method, params, data

    def update_dhcp_reservation(self, entity, hostname, macAddr, properties):
        if isinstance(entity['properties'], str):
            entity = self.prop_s2d(entity)
        old_properties = entity['properties']
        if isinstance(properties, str):
            properties = self.prop_s2d(properties)
        new_entity = entity.copy()
        new_entity['name'] = hostname
        properties['state'] = old_properties['state']
        properties['address'] = old_properties['address']
        properties['macAddress'] = macAddr
        properties['locationInherited'] = old_properties['locationInherited']
        new_entity['properties'] = self.prop_d2s(properties)
        self.update(new_entity)

    ################################################################
    # SEARCHES
    ################################################################

    @rest_call('get')
    def search_by_object_types(self, keyword, obj_type, start=0, count=100):
        method = 'searchByObjectTypes'
        params = {
            'keyword': keyword,
            'types': obj_type,
            'start': start,
            'count': count
        }
        data = None
        return method, params, data

    def search_ip_by_name(self, keyword, start=0, count=100):
        return self.search_by_object_types(keyword, 'IP4Address', start, count)

    ################################################################
    # NETWORK STUFF
    ################################################################

    @rest_call('get')
    def get_entity_by_cidr(self, parent_id, cidr, objType):
        """config.id only works for top-level blocks, parent_id must literally be the parent obect's id"""
        method = 'getEntityByCIDR'
        params = {
            'parentId': parent_id,
            'cidr': cidr,
            'type': objType
        }
        data = None
        return method, params, data

    def get_network_by_cidr(self, parent_id, cidr):
        return self.get_entity_by_cidr(parent_id, cidr, 'IP4Network')

    def get_block_by_cidr(self, parent_id, cidr):
        return self.get_entity_by_cidr(parent_id, cidr, 'IP4Block')

    @rest_call('get')
    def get_ip_ranged_by_ip(self, parentId, ipAddr, objType):
        method = 'getIPRangedByIP'
        params = {
            'containerId': parentId,
            'type': objType,
            'address': ipAddr.split('/')[0]
        }
        data = None
        return method, params, data

    def get_network(self, netAddr):
        return self.get_ip_ranged_by_ip(self.config['id'], netAddr, 'IP4Network')

    def get_network_by_ip(self, netAddr):
        return self.get_ip_ranged_by_ip(self.config['id'], netAddr, 'IP4Network')

    def get_block_by_ip(self, netAddr):
        return self.get_ip_ranged_by_ip(self.config['id'], netAddr, 'IP4Block')

    def get_dhcp_scope_by_ip(self, netAddr):
        return self.get_ip_ranged_by_ip(self.config['id'], netAddr, 'DHCP4Range')

    ################################################################
    # IP ADDRESS STUFF
    ################################################################

    @rest_call('get')
    def get_ip_address(self, ipAddr, parentId=None):
        method = 'getIP4Address'
        params = {
            'containerId': self.config['id'] if parentId is None else parentId,
            'address': ipAddr
        }
        data = None
        return method, params, data

    @rest_call('get')
    def get_next_ip_address(self, netId, offset=None, dhcpExclude=True):
        method = 'getNextIP4Address'
        properties = 'excludeDHCPRange={}|'.format(str(dhcpExclude).lower())
        if offset is not None:
            properties += 'offset={}|'.format(offset)
        params = {
            'parentId': netId,
            'properties': properties
        }
        data = None
        return method, params, data

    @rest_call('post')
    def assign_next_ip_address(self, parentId, hostname, macAddr=None, action='MAKE_STATIC', properties='', offset=None):
        method = 'assignNextAvailableIP4Address'
        if isinstance(properties, dict):
            properties = self.prop_d2s(properties)
        properties += 'name={}|'.format(hostname)
        if offset is not None:
            properties += 'offset={}|'.format(offset)
        params = {
            'configurationId': self.config['id'],
            'parentId': parentId,
            'macAddress': macAddr,
            'hostInfo': '',
            'action': action,
            'properties': properties
        }
        data = None
        return method, params, data

    @rest_call('post')
    def assign_ip_address(self, hostname, ipAddr, macAddr='', action='MAKE_STATIC', properties=''):
        method = 'assignIP4Address'
        if isinstance(properties, dict):
            properties = self.prop_d2s(properties)
        properties += 'name={}|'.format(hostname)
        params = {
            'configurationId': self.config['id'],
            'ip4Address': str(ipAddr),
            'macAddress': macAddr,
            'hostInfo': '',
            'action': action,
            'properties': properties
        }
        data = None
        return method, params, data

    def assign_ip_address_pair(self, net1, net2, hostname1, hostname2=None):
        """Assigns matching addresses in two separate, but equal-sized, networks
        networks expected to be CIDR"""
        if hostname2 is None:
            hostname2 = hostname1
        # Get Network Objects and set dhcp_offset based on env
        net1 = ip_network(unicode(net1))
        net2 = ip_network(unicode(net2))
        if net1.netmask != net2.netmask:
            raise ValueError('net1 and net2 CIDR prefixes are not equal')
        bam_net_1 = self.get_network(str(net1.network_address))
        bam_net_2 = self.get_network(str(net2.network_address))
        if net1.prefixlen > 24:
            dhcp_offset = 1
        else:
            dhcp_offset = 31
        mask = int(net1.netmask) ^ 2**32-1  # mask to determine ip's place in network
        # Ensure Hostname doesn't already exist in BAM
        found_ip = False
        response1 = self.get_entity_by_name(bam_net_1['id'], hostname1, 'IP4Address')
        response2 = self.get_entity_by_name(bam_net_2['id'], hostname2, 'IP4Address')
        if any(r['properties'] is not None for r in [response1, response2]):
            bam_ip_list = [self.entity_to_json(response1), self.entity_to_json(response2)]
            found_ip = True
        # If hostname doesn't exist, begin looping through networks for available address pairs
        while not found_ip and dhcp_offset < mask:
            dhcp_offset1 = str(net1.network_address + dhcp_offset)
            dhcp_offset2 = str(net2.network_address + dhcp_offset)
            ip1 = ip_address(self.get_next_ip_address(bam_net_1['id'], offset=dhcp_offset1))
            ip2 = ip_address(self.get_next_ip_address(bam_net_2['id'], offset=dhcp_offset2))
            if any(ip is None for ip in [ip1, ip2]):
                print('ERROR: out of IPs :(')
                found_ip = True
            else:
                # use mask to determine the ip's network index, e.g. in 10.20.30.0/23 10.20.31.15 = 271
                num1 = int(ip1) & int(mask)
                num2 = int(ip2) & int(mask)
                # if the ip's have the same network index, go ahead with assignment
                if num1 == num2:
                    ipObj1 = self.assign_ip_address(hostname1, str(ip1))
                    ipObj2 = self.assign_ip_address(hostname2, str(ip2))
                    bam_ip_list = [ipObj1, ipObj2]
                    found_ip = True
                # if the ip's have different indexes, continue looking at the highest of the two indexes
                elif num1 > num2:
                    dhcp_offset = num1 - 1
                else:
                    dhcp_offset = num2 - 1
        return bam_ip_list

################################################################
# MAC ADDRESS STUFF
################################################################

    @rest_call('get')
    def get_mac_address(self, mac):
        method = 'getMACAddress'
        params = {
            'configurationId': self.config['id'],
            'macAddress': mac.replace('.', '').replace(':', '').replace('-', '')
        }
        data = None
        return method, params, data

    @rest_call('post')
    def create_mac_address(self, mac, name, properties):
        method = 'addMACAddress'
        if isinstance(properties, dict):
            properties_string = 'name={}|'.format(name) + self.prop_d2s(properties)
        else:
            properties_string = 'name={}|'.format(name) + properties
        params = {
            'configurationId': self.config['id'],
            'macAddress': mac.replace('.', '').replace(':', '').replace('-', ''),
            'properties': properties_string
        }
        data = None
        return method, params, data

################################################################
# ROLE AND DEPLOYMENT STUFF
################################################################

    @rest_call('get')
    def get_deployment_roles(self, entity_id):
        method = 'getDeploymentRoles'
        params = {
            'entityId': entity_id
        }
        data = None
        return method, params, data

    @rest_call('get')
    def get_server_for_role(self, role_id):
        method = 'getServerForRole'
        params = {
            'roleId': role_id
        }
        data = None
        return method, params, data

    @rest_call('get')
    def get_deployment_status(self, server_id):
        method = 'getServerDeploymentStatus'
        params = {
            'serverId': server_id,
            'properties': ''
        }
        data = None
        return method, params, data

    @rest_call('post')
    def deploy_server(self, server_id):
        method = 'deployServer'
        params = {
            'serverId': server_id
        }
        data = None
        return method, params, data

    @rest_call('post')
    def deploy_server_config(self, server_id, services='DHCP', full=False):
        method = 'deployServerConfig'
        properties = 'ObjectProperties.services={}'.format(services)
        if full and 'DNS' in service:
            properties += '|forceDNSFullDeployment=true'
        params = {
            'serverId': server_id,
            'properties': properties
        }
        data = None
        return method, params, data

    @rest_call('post')
    def deploy_server_services(self, server_id, services='DHCP'):
        method = 'deployServerServices'
        params = {
            'serverId': server_id,
            'services': 'services={}'.format(services)
        }
        data = None
        return method, params, data

    def queue_servers(self, network_id, server_set=None):
        """Given a set and a network_id, determines the primary
        and backup servers for the network and adds them to the
        set. The resulting set is then returned.
        """
        if server_set is None:
            server_set = set()
        roles = self.get_deployment_roles(network_id)
        server_primary = self.get_server_for_role(roles[0]['id'])
        server_backup_id = DATA.ADONIS_PAIRS[server_primary['id']]
        server_set.add(server_primary['id'])
        server_set.add(server_backup_id)
        return server_set


    def deploy_dhcp_and_monitor(self, server_set):
        """Given a set of servers to deploy, each server will be
        deployed. The deployment status will be followed until all
        have terminated in some way.
        """
        for server in server_set:
            self.deploy_server_services(server, 'DHCP')
        self.monitor_server_deployment(server_set)


    def monitor_server_deployment(self, server_set):
        """Given a set of servers that have been deployed, the status
        of each will be continuously polled until all have reached a
        final state.
        """
        for server in server_set:
            status = self.get_deployment_status(server)
            self.logger.info('{} - {}'.format(DATA.ADONIS_ID_MAP[server], DATA.DEPLOYMENT_STATUS[status]))
        while len(server_set) > 0:
            sleep(2)
            servers = list(server_set)
            for server in servers:
                status = self.get_deployment_status(server)
                if status not in [-1, 0, 1]:
                    self.logger.info('{} - {}'.format(DATA.ADONIS_ID_MAP[server], DATA.DEPLOYMENT_STATUS[status]))
                    server_set.remove(server)

################################################################
# IF RUN DIRECTLY, MOSTLY FOR TESTING
################################################################

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('creds')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    args = parser.parse_args()
    # Load credentials
    with open(args.creds) as f:
        creds = json.load(f)
    # Enable logging if requested
    if args.loglevel:
        level = getattr(logging, args.loglevel.upper())
        logging.basicConfig(level=level)
    # Prove that client works
    c = BAM(creds['hostname'], creds['username'], creds['password'], loglevel=args.loglevel)
    net = c.get_network_by_ip('10.168.128.0')
    roles = c.get_deployment_roles(net['id'])
    r = c.get_next_ip_address(net['id'], offset='10.168.128.100')
    r = c.assign_next_ip_address(net['id'], 'offset-testing', offset='10.168.128.100')
    r = c.get_ip_address('10.97.12.69')
    c.logout()
    print(c.history)
    # print(data.ADONIS_PAIRS)




class BluecatError(Exception):
    def __init__(self, response):
        super(BluecatError, self).__init__(response.content)
        self.response = response
        self.status_code = response.status_code


class BluecatWebSystemError(BluecatError):
    pass




#!/usr/bin/python
import argparse
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def search_mac(session, mac):
    formatted_mac = mac.replace('.', '').replace(':', '').replace('-', '')
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('hostname', help='hostname')
    parser.add_argument('mac_addr', help='MAC Address')
    parser.add_argument('ip_addr', help='IP desired for reservation, also defines the network if "first_available" flag is set')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    parser.add_argument('--location', default='', help='Location of Device')
    parser.add_argument('--notes', default='', help='Additional notes about the reservation')
    parser.add_argument('--owner', default='', help='Owner of the device')
    parser.add_argument('--request_num', default='', help='ServiceDesk Request Number')
    parser.add_argument('--no_wait', action='store_true', help='do NOT wait for deployment')
    parser.add_argument('--no_deploy', action='store_true', help='Do NOT deploy saved changes')
    args = parser.parse_args()

    # Setup console logging
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(level=loglevel)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.propagate = False

    # Setup args and vars
    hostname = args.hostname
    mac = args.mac_addr.replace('.', '').replace(':', '').replace('-', '')
    ip = ip_address(unicode(args.ip_addr))
    action = 'MAKE_DHCP_RESERVED'
    creds = pybluecat.get_creds(args.creds)
    properties = {
        'Location': args.location,
        'Notes': args.notes,
        'Owner': args.owner,
        'RequestNum': args.request_num
    }
    output_object = {}

    # Initialize flow-control boolean vars
    deploy_needed = False
    ip_is_eligible = False
    mac_already_reserved = False
    desired_ip_reserved = False
    error_message = ''

    # Create instance using 'with' so cleanup of session is automatic
    with pybluecat.BAM(**creds) as bam:
        # Determine the Network of the IP Address
        try:
            logger.info('Getting Network info from ip: {}'.format(str(ip)))
            net_entity = bam.get_network(str(ip))
            net_obj = pybluecat.entity_to_json(net_entity)
            network = ip_network(unicode(net_obj['properties']['CIDR']))
            if network.prefixlen <= 24:
                dhcp_offset_ip = str(network.network_address + 31)
            else:
                dhcp_offset_ip = None
            logger.debug(json.dumps(net_obj, indent=2))
        except pybluecat.exceptions.BluecatError as e:
            logger.error('Could not determine the target network or network did not exist')
            exit(str(e))

        # Determine if IP eligible to be reserved, not Broadcast, Network, Gateway, etc...
        if ip == network.network_address:
            error_message = 'Cannot Update the Network Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip == network.broadcast_address:
            error_message = 'Cannot Update the Broadcast Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip == network.network_address + 1:
            error_message = 'Cannot Update the Gateway Address: {}'.format(str(ip))
            logger.info(error_message)
        elif ip in [network.network_address + 2, network.network_address + 3]:
            error_message = 'Cannot Update a HSRP/VRRP Address: {}'.format(str(ip))
            logger.info(error_message)
        elif network.prefixlen <= 24 and ip in [network.network_address + i for i in xrange(4, 31)]:
            error_message = 'Cannot Update Address in Telecom-Reserved Space: {}'.format(str(ip))
            logger.info(error_message)
        else:
            ip_is_eligible = True


        # Assuming the IP is eligible, attempt the update
        if ip_is_eligible:
            try:
                logger.info('Getting current config for IP: {}'.format(str(ip)))
                ip_entity = bam.get_ip_address(str(ip))
                old_reservation = pybluecat.entity_to_json(ip_entity)
                logger.debug(json.dumps(old_reservation, indent=2, sort_keys=True))
                if old_reservation['properties'] is None:
                    output_object = {
                        'state': 'No Update Performed',
                        'message': 'Target IP is not a DHCP_RESERVATION',
                        'existing_reservation': old_reservation
                    }
                elif old_reservation['properties']['state'] != 'DHCP_RESERVED':
                    output_object = {
                        'state': 'No Update Performed',
                        'message': 'Target IP is not a DHCP_RESERVATION',
                        'existing_reservation': old_reservation
                    }
                else:
                    response = bam.update_dhcp_reservation(old_reservation, hostname, mac, properties)
                    new_reservation = bam.get_ip_address(str(ip))
                    new_reservation = pybluecat.entity_to_json(new_reservation)
                    logger.debug(json.dumps(old_reservation, indent=2, sort_keys=True))
                    output_object = {
                        'status': 'Reservation Updated Successfully',
                        'message': 'Target IP was updated',
                        'old_reservation': old_reservation,
                        'new_reservation': new_reservation
                    }
                    deploy_needed = True
            except pybluecat.exceptions.BluecatError as e:
                output_object = {
                    'status': 'Error Occurred during Update',
                    'message': e.message
                }
        else:
            output_object = {
                'status': 'No Update Performed',
                'message': error_message
            }

        # Deploy DHCP changes if no_deploy flag isn't set
        if deploy_needed and not args.no_deploy:
            logger.info('Deploying Config changes')
            server_set = bam.queue_servers(net_obj['id'])
            for server in server_set:
                logger.info('Deploying changes to server: {}'.format(str(server)))
                bam.deploy_server_services(server, 'DHCP')
            # Monitor the deployment through completion unless no_wait flag is set
            if not args.no_wait:
                logger.info('Waiting for deployment to finish')
                bam.monitor_server_deployment(server_set)
                logger.info('Deployment finished')
            else:
                logger.info('Monitoring of deployment is being skipped')

    # Finally output the IP info
    #print_ip_info(ip_obj)
    print(json.dumps(output_object, indent=2, sort_keys=True))


if __name__ == '__main__':
    main()




#!/usr/bin/python
from pybluecat.data.helpers import *


class APIEntity(object):

    def __init__(id, type, name=None, value=None, properties=None):
        self.id = id
        self.type = type
        self.name = name
        self.value = value
        self.properties = properties_s2d(properties)

    def __str__(self):
        pass



import pybluecat.data.const
import pybluecat.data.helpers
from pybluecat.data.defs import *


DEPLOYMENT_STATUS = {
    -1: 'EXECUTING',
    0: 'INITIALIZING',
    1: 'QUEUED',
    2: 'CANCELLED',
    3: 'FAILED',
    4: 'NOT_DEPLOYED',
    5: 'WARNING',
    6: 'INVALID',
    7: 'DONE',
    8: 'NO_RECENT_DEPLOYMENT'
}

# Adonis Server ID's
bl_main = 557447
bdc_main = 5153278
tis_main = 557077
bl_cache = 1409300
bw_cache = 1409298
gmh_dc = 3496237
gmh_mdf = 3497381
sbr_dc = 3620429
sbr_mdf = 3620426
slh_dc = 3684293
slh_mdf = 3684295
zch_dc = 3548376
zch_2069 = 3549505
shp_dc = 3722273
shp_mdf = 3722278

ADONIS_ID_MAP = {
    557447: 'BL-MAIN',
    5153278: 'BDC-NA-DHCPDNS-DC-1',
    557077: 'TIS-NA-DHCPDNS-DC-1',
    1409300: 'BL-NA-DNSCACHE-DC-1',
    1409298: 'BW-NA-DNSCACHE-DC-1',
    3496237: 'gmh-na-dhcpdns-dc-1',
    3497381: 'gmh-na-dhcpdns-2036a-1',
    3620429: 'sbr-na-dhcpdns-dc-1',
    3620426: 'sbr-na-dhcpdns-mdf-1',
    3684293: 'slh-na-dhcp-dc-1',
    3684295: 'slh-na-dhcp-mdf-1',
    3548376: 'zch-na-dhcp-dc-1',
    3549505: 'zch-na-dhcp-2069-1',
    3722273: 'shp-na-dhcp-dc-1',
    3722278: 'shp-na-dhcp-mdf-1'
}

ADONIS_PAIRS = {
    bdc_main: tis_main,
    tis_main: bdc_main,
    bl_cache: bw_cache,
    bw_cache: bl_cache,
    gmh_dc: gmh_mdf,
    gmh_mdf: gmh_dc,
    sbr_dc: sbr_mdf,
    sbr_mdf: sbr_dc,
    slh_dc: slh_mdf,
    slh_mdf: slh_dc,
    zch_dc: zch_2069,
    zch_2069: zch_dc,
    shp_dc: shp_mdf,
    shp_mdf: shp_dc
}



#!/usr/bin/python
import argparse
import json
import logging
import os
import pybluecat
from ipaddress import ip_address, ip_network
from shutil import copyfile
from sys import exit


def search_ip(session, ip):
    results = session.get_ip_address(ip)
    return [results]


def search_mac(session, mac):
    formatted_mac = mac.replace('.', '').replace(':', '').replace('-', '')
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results


def search_name(session, name):
    results = session.search_ip_by_name(name)
    return results


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    group_me_in = parser.add_mutually_exclusive_group(required=True)
    group_me_in.add_argument('-i', '--ip', help='IP address')
    group_me_in.add_argument('-m', '--mac', help='MAC address')
    group_me_in.add_argument('-n', '--name', help='Device name')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    group_me_out = parser.add_mutually_exclusive_group()
    group_me_out.add_argument('--text', help='text output', action='store_true')
    group_me_out.add_argument('--csv', help='csv output', action='store_true')
    group_me_out.add_argument('--json', help='json output, this is the default', action='store_true')
    args = parser.parse_args()

    # Handle loading of credentials
    creds = pybluecat.get_creds(args.creds)
    hostname = creds['hostname']
    username = creds['username']
    password = creds['password']

    # Enable logging if requested
    # if args.loglevel:
    #     level = getattr(logging, args.loglevel.upper())
    #     logging.basicConfig(level=level)

    # Instantiate Bluecat REST Client
    client = pybluecat.BAM(hostname, username, password, loglevel=args.loglevel)

    if args.ip:
        results = search_ip(client, args.ip)
    elif args.mac:
        results = search_mac(client, args.mac)
    elif args.name:
        results = search_name(client, args.name)

    if args.text:
        for result in results:
            result['properties'] = client.prop_s2d(result['properties'])
            if args.text:
                print('Name       : {}'.format(result['name']))
                print('State      : {}'.format(result['properties']['state']))
                print('IP Address : {}'.format(result['properties']['address']))
                if 'macAddress' in result['properties']:
                    print('MAC Address: {}'.format(result['properties']['macAddress']))
                print('\n\n')
    elif args.csv:
        print('Name,State,IP,MAC')
        for result in results:
            result['properties'] = client.prop_s2d(result['properties'])
            name = result['name']
            state = result['properties']['state']
            ip = result['properties']['address']
            if 'macAddress' in result['properties']:
                mac = result['properties']['macAddress']
            else:
                mac = 'N/A'
            print(','.join([name, state, ip, mac]))
    else:
        print(json.dumps(results, indent=2, sort_keys=True))

if __name__ == "__main__":
    main()



#!/usr/bin/python
"""Walk BAM and output all subnets and associated info
"""
from ipaddress import ip_address, ip_network
import argparse
import json
import pybluecat
import re
import sys


CIDR_REGEX = re.compile(r'(?<=CIDR=)[^|]+')


def block_in_blacklist(block):
    try:
        cidr = CIDR_REGEX.findall(block['properties'])[0]
    except Exception:
        print(json.dumps(block, indent=2))
        sys.exit('Couldn\'t find CIDR in Block Properties')
    return cidr in CIDR_BLACKLIST


def name_in_list(entity, kw_list):
    if entity['name'] is None:
        entity['name'] = ''
    return any(keyword in entity['name'].lower() for keyword in kw_list)


def get_line(session, netObj):
    properties = session.prop_s2d(netObj['properties'])
    cidr = properties['CIDR'] if 'CIDR' in properties else ''
    vlan = properties['Vlan'] if 'Vlan' in properties else ''
    loc = properties['Location'] if 'Location' in properties else ''
    return '{c},"{n}",{t},{v},"{l}"'.format(c=cidr, n=netObj['name'], t=netObj['type'], v=vlan, l=loc)


def walk_networks(session, start):
    # Get all Blocks and Subnets within the Start Block
    blocks = session.get_entities(start['id'], 'IP4Block', 0, 256)
    subnets = session.get_entities(start['id'], 'IP4Network', 0, 256)
    # If there are Blocks within Start, evaluate them
    if len(blocks) > 0:
        for block in blocks:
            # Do any filters preclude this block from use?
            if (block_in_blacklist(block)
                or name_in_list(block, KW_BLACKLIST)
                or not name_in_list(block, KW_WHITELIST)):
                # Skip block
                continue
            # Print the Block info if Global flag not set
            if not IGNORE_BLOCKS:
                print(get_line(session, block))
            # RECURSE!!!
            walk_networks(session, block)
    # If there are Subnets within Start, evaluate them
    for subnet in subnets:
        # Do any filters preclude this subnet from use?
        if (name_in_list(subnet, KW_BLACKLIST)
            or not name_in_list(subnet, KW_WHITELIST)):
            # Skip subnet
            continue
        # searches properties string for the CIDR, then creates net obj
        net = ip_network(unicode(CIDR_REGEX.findall(subnet['properties'])[0]))
        # print the network if it meets size requirements
        if (net.prefixlen <= MINIMUM
            and net.prefixlen >= MAXIMUM):
            print(get_line(session, subnet))


def find_start_block(session, cidr, start_id):
    cidr = ip_network(unicode(cidr))
    top_blocks = session.get_entities(start_id, 'IP4Block', 0, 1000)
    for block in top_blocks:
        block_cidr = pybluecat.prop_s2d(block['properties'])['CIDR']
        block_net = ip_network(block_cidr)
        if block_net.overlaps(cidr):
            if block_net == cidr:
                return block
            else:
                return(find_start_block(session, cidr, block['id']))


def main():
    # Get and Parse all CLI arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('cidr', nargs='+', help='CIDR(s) to act as root for enumeration')
    parser.add_argument('-a', '--append', help='omit header for appending to previous output', action='store_true')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('--minimum', type=int, default=32, help='Minimum size (inclusive) of subnets, e.g. "24"')
    parser.add_argument('--maximum', type=int, default=0, help='Maximum size (inclusive) of subnets, e.g. "24"')
    parser.add_argument('--ignore_blocks', action='store_true', help='omits IP4Blocks from output')
    parser.add_argument('--cidr_blacklist', help='filepath to list of network blocks to not enumerate')
    kw_group = parser.add_mutually_exclusive_group()
    kw_group.add_argument('--keyword_blacklist', help='filepath to list of name keywords to not enumerate')
    kw_group.add_argument('--keyword_whitelist', help='filepath to list of name keywords to not enumerate')
    args = parser.parse_args()

    # Initialize the Global Ignore Blocks flag
    global IGNORE_BLOCKS
    IGNORE_BLOCKS = args.ignore_blocks
    # Initialize the Global CIDR Minimum size
    global MINIMUM
    MINIMUM = args.minimum
    # Initialize the Global CIDR Maximum size
    global MAXIMUM
    MAXIMUM = args.maximum
    # Initialize the Global CIDR Blacklist
    global CIDR_BLACKLIST
    if args.cidr_blacklist is not None:
        try:
            with open(args.cidr_blacklist) as f:
                CIDR_BLACKLIST = [line.strip().split()[0] for line in f.readlines() if line.strip() != '']
        except Exception:
            sys.exit('blacklist {} could not be parsed'.format(args.cidr_blacklist))
    else:
        CIDR_BLACKLIST = []
    # Initialize the Global Keyword Blacklist
    global KW_BLACKLIST
    if args.keyword_blacklist is not None:
        try:
            with open(args.keyword_blacklist) as f:
                KW_BLACKLIST = [line.strip().split()[0].lower() for line in f.readlines() if line.strip() != '']
        except Exception:
            sys.exit('blacklist {} could not be parsed'.format(args.keyword_blacklist))
    else:
        KW_BLACKLIST = []
    # Initialize the Global Keyword Whitelist
    global KW_WHITELIST
    if args.keyword_whitelist is not None:
        try:
            with open(args.keyword_whitelist) as f:
                KW_WHITELIST = [line.strip().split()[0].lower() for line in f.readlines() if line.strip() != '']
        except Exception:
            sys.exit('whitelist {} could not be parsed'.format(args.keyword_blacklist))
    else:
        KW_WHITELIST = ['']

    # Get Credentials
    creds = pybluecat.get_creds(args.creds)
    # Create the Bluecat sessions
    with pybluecat.BAM(creds['hostname'], creds['username'], creds['password']) as c:
        config = c.get_entity_by_name(0, 'Spectrum Health', 'Configuration')
        # Output headers if append flag not set
        if not args.append:
            print('Network,Name,Type,VLAN,Location')
        # Walk Bluecat from each starting block
        for cidr in args.cidr:
            startBlock = find_start_block(c, cidr, config['id'])
            if not IGNORE_BLOCKS:
                print(get_line(c, startBlock))
            walk_networks(c, startBlock)


if __name__ == '__main__':
    main()



#!/usr/bin/python


def properties_s2d(self, propString):
    return {p[0]: p[1] for p in [pair.split('=') for pair in propString.split('|')[:-1]]}


def properties_d2s(self, propDict):
    return '|'.join(['='.join(pair) for pair in propDict.items()]) + '|'



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()

setup(
        name='pybluecat',
        version='0.1.1',
        description='Python wrapper around Bluecat APIs',
        long_description=readme(),
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.6',
            'License :: Freely Distributable',
            'Natural Language :: English',
        ],
    url='https://github.com/ForrestT/pyBluecat',
        author='Forrest Throesch',
        author_email='fmthroesch@gmail.com',
        license='Freely Distributable',
        packages=[
            'pybluecat',
            'pybluecat.tools',
            'pybluecat.data'
        ],
        entry_points={
            'console_scripts': [
                'bluecat=pybluecat.tools.cli:main',
                'bluecat-assinv-search=pybluecat.tools.assinv_search:main',
                'bluecat-dhcp-request=pybluecat.tools.rundeck_dhcp_request:main',
                'bluecat-dhcp-bulk-request=pybluecat.tools.rundeck_dhcp_bulk_request:main',
                'bluecat-dhcp-update=pybluecat.tools.rundeck_dhcp_update:main',
                'bluecat-dhcp-delete=pybluecat.tools.rundeck_dhcp_delete:main',
                'bluecat-dhcp-search=pybluecat.tools.rundeck_dhcp_search:main',
                'bluecat-vip-request=pybluecat.tools.vips:main',
                'bluecat-search=pybluecat.tools.search:main',
                'bluecat-network-enumerate=pybluecat.tools.enumerate_networks:main',
                'bluecat-static-request=pybluecat.tools.static_request:main'
            ]
        },
        install_requires=[
            'dnspython'
        ],
        zip_safe=False)



#!/usr/bin/python
import argparse
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def ping(host):
    import subprocess
    try:
        r = subprocess.check_output(['ping', '-c', '1', host], stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False

def dns_A_exists(host):
    try:
        a = resolver.query(host + '.spectrum-health.org', 'A')
        return a
    except:
        return None

def dns_PTR_exists(host):
    try:
        ptr = reversename.from_address(host)
        name = resolver.query(ptr, 'PTR')[0]
        return str(name)
    except:
        return None

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('hostname', help='hostname')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    group_ME = parser.add_mutually_exclusive_group(required=True)
    group_ME.add_argument('-n', '--network', nargs=2, help='network address within desired subnet')
    group_ME.add_argument('-e', '--environment', help='environment from which to choose a network')
    parser.add_argument('-l', '--loglevel', choices=['warning', 'info', 'debug'], help='enable debugging')
    args = parser.parse_args()

    NETWORK_ENVS = {
        'lab': [u'10.168.131.0/24', u'10.168.161.0/24'],
        'dmz-vip': [u'167.73.15.0/24', u'167.73.31.0/24']
        }

    if args.loglevel:
        level = getattr(logging, args.loglevel.upper())
        logging.basicConfig(level=level)

    hostname = args.hostname.lower()
    creds = pybluecat.get_creds(args.creds)
    bam = pybluecat.BAM(creds['hostname'], creds['username'], creds['password'])

    # Get Networks List
    if args.network:
        net1 = args.network[0]
        net2 = args.network[1]
    elif args.environment:
        net1, net2 = NETWORK_ENVS[args.environment]

    # Reserve the pairs
    addresses = bam.assign_ip_address_pair(net1, net2, args.hostname)
    print(json.dumps(addresses, indent=2))


if __name__ == '__main__':
    main()




#!/usr/bin/python
import argparse
import json
import logging
import os
import pybluecat
from ipaddress import ip_address, ip_network
from shutil import copyfile
from sys import exit


def search_ip(session, ip):
    ip_entity = session.get_ip_address(ip)
    ip = session.entity_to_json(ip_entity)
    if 'macAddress' in ip['properties']:
        mac_entity = session.get_mac_address(ip['properties']['macAddress'])
        mac = session.entity_to_json(mac_entity)
        output = {
            'ip': ip,
            'mac': mac
        }
    else:
        output = {
            'ip': ip,
            'mac': None
        }
    return output


def search_mac(session, mac):
    formatted_mac = mac.replace('.', '').replace(':', '').replace('-', '')
    mac_entity = session.get_mac_address(formatted_mac)
    linked_ips = session.get_linked_entities(mac_entity['id'])
    linked_ips = [session.entity_to_json(linked_ip) for linked_ip in linked_ips]
    output = {
        'mac': session.entity_to_json(mac_entity),
        'linked_ips': linked_ips
    }
    return output


def search_name(session, name):
    results = session.search_ip_by_name(name)
    return results


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-m', '--mac', help='MAC address')
    parser.add_argument('-n', '--name', help='Device name')
    parser.add_argument('-p', '--properties', help='Properties')
    parser.add_argument('-l', '--loglevel', choices=['critical', 'error', 'warning', 'info', 'debug'],
                        default='critical', help='enable logging')
    args = parser.parse_args()

    # Handle loading of credentials
    creds = pybluecat.get_creds(args.creds)
    hostname = creds['hostname']
    username = creds['username']
    password = creds['password']

    # Enable logging if requested
    # if args.loglevel:
    #     level = getattr(logging, args.loglevel.upper())
    #     logging.basicConfig(level=level)

    # Instantiate Bluecat REST Client
    bam = pybluecat.BAM(hostname, username, password, loglevel=args.loglevel)
    r = bam.create_mac_address(args.mac, args.name, args.properties)
    print(json.dumps(r, indent=2))


if __name__ == "__main__":
    main()



#!/usr/bin/python
import argparse
import json
import logging
import os
import pybluecat
from ipaddress import ip_address, ip_network
from shutil import copyfile
from sys import exit


def search_ip(session, ip):
    results = session.get_ip_address(ip)
    return [results]


def format_mac(mac):
    return mac.replace('.', '').replace(':', '').replace('-', '').lower()


def search_mac(session, mac):
    formatted_mac = format_mac(mac)
    mac_entity = session.get_mac_address(formatted_mac)
    results = session.get_linked_entities(mac_entity['id'])
    return results


def search_name(session, name):
    results = session.search_ip_by_name(name)
    return results


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    parser.add_argument('-i', '--ip', help='IP address')
    parser.add_argument('-m', '--mac', help='MAC address')
    parser.add_argument('-n', '--name', help='Device name')
    parser.add_argument('-l', '--loglevel', default='critical', help='enable logging',
                        choices=['critical', 'error', 'warning', 'info', 'debug'])
    match_group = parser.add_mutually_exclusive_group()
    match_group.add_argument('--match_all', action='store_true')
    match_group.add_argument('--match_any', action='store_true')
    args = parser.parse_args()

    # Setup console logging
    logger = logging.getLogger(__name__)
    loglevel = getattr(logging, args.loglevel.upper())
    logger.setLevel(level=loglevel)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.propagate = False

    # Handle loading of credentials
    creds = pybluecat.get_creds(args.creds)
    hostname = creds['hostname']
    username = creds['username']
    password = creds['password']

    # Instantiate Bluecat REST Client
    bam = pybluecat.BAM(hostname, username, password)  # , loglevel=args.loglevel)

    # Normalizing Arguments due to Rundeck requirements
    if args.ip == '':
        ip = None
    else:
        ip = args.ip
    if args.mac == '':
        mac = None
    else:
        mac = args.mac
    if args.name == '':
        name = None
    else:
        name = args.name

    # Use provided fields to search Bluecat Objects 
    results = []
    if ip is not None:
        results += search_ip(bam, ip)
    if mac is not None:
        results += search_mac(bam, mac)
    if name is not None:
        results += search_name(bam, name)

    # Eliminate Dupes and conflicting filters
    matches = []
    for result in results:
        ip_obj = bam.entity_to_json(result)
        # Skip NULL objects
        if ip_obj['properties'] is None:
            continue
        # Skip Duplicates
        if ip_obj in matches:
            continue
        # Gather details about the ip_object
        res_name = ip_obj['name']
        res_ip = ip_obj['properties']['address']
        if 'macAddress' in ip_obj['properties']:
            res_mac = ip_obj['properties']['macAddress']
        else:
            res_mac = ''
        logger.info('NAME: {}, IP: {}, MAC: {}'.format(res_name, res_ip, res_mac))
        # If IP Object state is not "DHCP_RESERVED" DO NOT DELETE
        if ip_obj['properties']['state'] != 'DHCP_RESERVED':
            logger.info('IP Object is NOT a DHCP Reservation, ignoring objects with state: {}'.format(ip_obj['properties']['state']))
            continue
        # If --match_all is set, ensure ALL set fields match the reservation
        if args.match_all:
            # If --hostname is set and doesn't match, skip to next reservation
            if name is not None and res_name.lower() != name.lower():
                logger.info('Reservation doesn\'t match --hostname filter with --match_all set, ignoring')
                continue
            # If --ip_addr is set and doesn't match, skip to next reservation
            if ip is not None and res_ip != str(ip):
                logger.info('Reservation doesn\'t match --ip_addr filter with --match_all set, ignoring')
                continue
            # If --mac_addr is set and doesn't match, skip to next reservation
            if mac is not None and format_mac(res_mac) != mac:
                logger.info('Reservation doesn\'t match --mac_addr filter with --match_all set, ignoring')
                continue
        matches.append(ip_obj)

    # Output the remaining matched ip objects
    print(json.dumps(matches, indent=2, sort_keys=True))


if __name__ == "__main__":
    main()




#!/usr/bin/python
import argparse
import json
import logging
import pybluecat
from ipaddress import ip_address, ip_network
from sys import exit


def ping(host):
    import subprocess
    try:
        r = subprocess.check_output(['ping', '-c', '1', host], stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False

def dns_A_exists(host):
    try:
        a = resolver.query(host + '.spectrum-health.org', 'A')
        return a
    except:
        return None

def dns_PTR_exists(host):
    try:
        ptr = reversename.from_address(host)
        name = resolver.query(ptr, 'PTR')[0]
        return str(name)
    except:
        return None

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('hostname', help='hostname')
    parser.add_argument('-c', '--creds', help='path to file containing credentials')
    group_ME = parser.add_mutually_exclusive_group(required=True)
    group_ME.add_argument('-n', '--network', help='network address within desired subnet')
    group_ME.add_argument('-e', '--environment', help='environment from which to choose a network')
    parser.add_argument('-l', '--loglevel', choices=['warning', 'info', 'debug'], help='enable debugging')
    args = parser.parse_args()

    NETWORK_ENVS = {
        'lab': [ip_network(u'10.168.131.0/24'), ip_network(u'10.168.161.0/24')],
        'dev': [ip_network(u'10.57.128.0/23')],
        'test': [ip_network(u'10.57.144.0/23')],
        'stage': [ip_network(u'10.57.160.0/23')],
        'prod-ctis': [ip_network(u'10.7.96.0/23'), ip_network(u'10.7.98.0/23')],
        'prod-brad': [ip_network(u'10.107.96.0/23'), ip_network(u'10.107.98.0/23')],
        'prod-both': [ip_network(u'10.200.112.0/24'), ip_network(u'10.200.113.0/24')],
        'fmt': [ip_network(u'10.57.136.0/30'), ip_network(u'10.57.136.4/30')]
        }

    if args.loglevel:
        level = getattr(logging, args.loglevel.upper())
        logging.basicConfig(level=level)

    hostname = args.hostname.lower()
    creds = pybluecat.get_creds(args.creds)
    c = pybluecat.BAM(creds['hostname'], creds['username'], creds['password'])

    # Get Networks List
    if args.network:
        netAddr = args.network.split('/')[0]
        netObj = c.get_network(netAddr)
        netObj = c.entity_to_json(netObj)
        networks = [ip_network(netObj['properties']['CIDR'])]
    elif args.environment:
        import requests
        base_url = 'http://infradevapi.spectrum-health.org/infrastructure/'
        env = args.environment.lower()
        url = base_url + env
        proxy = {'http': None, 'https': None}
        if 'dmz' in args.environment:
            dhcp_offset = 0
        try:
            r = requests.get(url, proxies=proxy)
            logging.info(r.request.url)
            logging.info(r.status_code)
            logging.info(r.content)
        except:
            print('Environment Not Found. No Network Available.')
            exit()
        networks = [
            ip_network(net['netaddr'] + '/' + net['netmask'], strict=False)
            for net in json.loads(r.text)[env]
        ]
    logging.info('Networks: {}'.format(str(networks)))


    foundIP = False
    # Check for existing IP reservations in target networks
    results = c.search_by_object_types(hostname, 'IP4Address', 0, 1000)
    logging.info(str(results))
    for result in results:
        temp = c.entity_to_json(result)
        for net in networks:
            if ip_address(temp['properties']['address']) in net:
                foundIP = True
                ipObj = temp
                networks = [net]
                logging.info('Found IP already in existence: {}'.format(json.dumps(ipObj, indent=2)))
                break
    for network in networks:
        logging.info('Working through network: {}'.format(str(network)))
        # Get Network Object and set dhcp_offset based on CIDR
        netObj = c.get_network(str(network.network_address))
        netObj = c.entity_to_json(netObj)
        if network.prefixlen > 24:
            dhcp_offset = 0
        else:
            dhcp_offset = 30
        logging.info('DHCP-Offsest: {}'.format(str(dhcp_offset)))
        # Ensure IPs in the offset are 'reserved'
        while not foundIP:
            logging.info('Checking Status of Offset Addreses')
            ip = c.get_next_ip_address(netObj['id'])
            logging.info('Address {}'.format(str(ip)))
            if ip is None or ip == '':
                break
            elif network.network_address + dhcp_offset >= ip_address(unicode(ip)):
                c.assign_ip_address('', ip, '', 'MAKE_RESERVED', '')
                logging.info('Setting IP Address as RESERVED: {}'.format(ip))
            else:
                break
        # If an existing IP has not been found yet, start working through
        # every free IP in the BAM Network until one is assigned or net is
        # exhausted
        if not foundIP:
            logging.info('Determining next available IP Address')
            while True:
                ipObj = c.assign_next_ip_address(netObj['id'], hostname)
                # None as a result indicates network has no next IP, end loop
                if ipObj['id'] == 0:
                    break
                ipObj = c.entity_to_json(ipObj)
                logging.info('IP Address free in BAM: {}'.format(ipObj['properties']['address']))
                # Check if IP has existing PTR record, if True, write it to BAM, try next IP
                ptr = dns_PTR_exists(ipObj['properties']['address'])
                if ptr:
                    logging.info('PTR Record found for Address: {}'.format(ptr))
                    ipObj['name'] = ptr.split('.')[0]
                    ipObj = c.json_to_entity(ipObj)
                    c.update(ipObj)
                # Try to Ping the IP address, if response, log in BAM, try next IP
                elif ping(ipObj['properties']['address']):
                    logging.info('Address responded to ping')
                    ipObj['name'] = 'IN-USE: something pinged'
                    ipObj = c.json_to_entity(ipObj)
                    c.update(ipObj)
                # Finally, reserve the IP in BAM for the hostname
                else:
                    logging.info('Address doesn\'t ping or have PTR record')
                    foundIP = True
                    break
        # If an IP has been found, either new or existing, return results and exit
        if foundIP:
            network = ip_network(unicode(netObj['properties']['CIDR']))
            output = {
                'ip_addr': ipObj['properties']['address'],
                'gateway': str(network.network_address + 1),
                'net_mask': str(network.netmask),
                'net_name': netObj['name'],
                '_ipobj': ipObj,
                '_netobj': netObj
            }
            print(json.dumps(output, sort_keys=True, indent=4))
            c.logout()
            exit()
    if not foundIP:
        print('No Addresses Available.')
        exit(1)


if __name__ == '__main__':
    main()

